@@ -9,7 +9,10 @@
     (at your option) any later version.  See <https://www.gnu.org/licenses/>.
 */
 
+#include "fmpz_poly.h"
 #include "fmpz_poly_factor.h"
+#include "fmpzi.h"
+#include "arb_fmpz_poly.h"
 #include "acb.h"
 #include "acb_poly.h"
 #include "acb_mat.h"
@@ -19,21 +22,24 @@
 #include "acb_modular.h"
 #include "acb_elliptic.h"
 #include "acf.h"
-#include "fmpzi.h"
 #include "qqbar.h"
-#include "arb_fmpz_poly.h"
 #include "gr.h"
 #include "gr_generic.h"
 #include "gr_vec.h"
 #include "gr_poly.h"
+#include "nfloat.h"
 
 typedef struct
 {
     slong prec;
+    int flags;
 }
 gr_acb_ctx;
 
+#define ACB_CTX_ANALYTIC 1
+
 #define ACB_CTX_PREC(ring_ctx) (((gr_acb_ctx *)((ring_ctx)))->prec)
+#define ACB_CTX_FLAGS(ring_ctx) (((gr_acb_ctx *)((ring_ctx)))->flags)
 
 int _gr_acb_ctx_set_real_prec(gr_ctx_t ctx, slong prec)
 {
@@ -276,6 +282,18 @@ _gr_acb_set_other(acb_t res, gr_srcptr x, gr_ctx_t x_ctx, gr_ctx_t ctx)
                 return GR_DOMAIN;
             }
 
+        case GR_CTX_NFLOAT_COMPLEX:
+            if (NFLOAT_CTX_HAS_INF_NAN(x_ctx)) /* todo */
+            {
+                return GR_UNABLE;
+            }
+            else
+            {
+                nfloat_complex_get_acb(res, x, x_ctx);
+                acb_set_round(res, res, ACB_CTX_PREC(ctx));
+                return GR_SUCCESS;
+            }
+
         case GR_CTX_RR_ARB:
             arb_set_round(acb_realref(res), x, ACB_CTX_PREC(ctx));
             arb_zero(acb_imagref(res));
@@ -722,7 +740,10 @@ _gr_acb_pow(acb_t res, const acb_t x, const acb_t exp, const gr_ctx_t ctx)
     }
     else if (!acb_contains_zero(x) || arb_is_positive(acb_realref(exp)))
     {
-        acb_pow(res, x, exp, ACB_CTX_PREC(ctx));
+        if (ACB_CTX_FLAGS(ctx) & ACB_CTX_ANALYTIC)
+            acb_pow_analytic(res, x, exp, 1, ACB_CTX_PREC(ctx));
+        else
+            acb_pow(res, x, exp, ACB_CTX_PREC(ctx));
 
         if (!acb_is_finite(res))
             return GR_UNABLE;
@@ -760,7 +781,17 @@ _gr_acb_is_square(const acb_t x, const gr_ctx_t ctx)
 int
 _gr_acb_sqrt(acb_t res, const acb_t x, const gr_ctx_t ctx)
 {
-    acb_sqrt(res, x, ACB_CTX_PREC(ctx));
+    if (ACB_CTX_FLAGS(ctx) & ACB_CTX_ANALYTIC)
+    {
+        acb_sqrt_analytic(res, x, 1, ACB_CTX_PREC(ctx));
+        if (!acb_is_finite(res))
+            return GR_UNABLE;
+    }
+    else
+    {
+        acb_sqrt(res, x, ACB_CTX_PREC(ctx));
+    }
+
     return GR_SUCCESS;
 }
 
@@ -769,7 +800,17 @@ _gr_acb_rsqrt(acb_t res, const acb_t x, const gr_ctx_t ctx)
 {
     if (!acb_contains_zero(x))
     {
-        acb_rsqrt(res, x, ACB_CTX_PREC(ctx));
+        if (ACB_CTX_FLAGS(ctx) & ACB_CTX_ANALYTIC)
+        {
+            acb_rsqrt_analytic(res, x, 1, ACB_CTX_PREC(ctx));
+            if (!acb_is_finite(res))
+                return GR_UNABLE;
+        }
+        else
+        {
+            acb_rsqrt(res, x, ACB_CTX_PREC(ctx));
+        }
+
         return GR_SUCCESS;
     }
     else if (acb_is_zero(x))
@@ -785,16 +826,36 @@ _gr_acb_rsqrt(acb_t res, const acb_t x, const gr_ctx_t ctx)
 int
 _gr_acb_floor(acb_t res, const acb_t x, const gr_ctx_t ctx)
 {
-    arb_floor(acb_realref(res), acb_realref(x), ACB_CTX_PREC(ctx));
-    arb_zero(acb_imagref(res));
+    if (ACB_CTX_FLAGS(ctx) & ACB_CTX_ANALYTIC)
+    {
+        acb_real_floor(res, x, 1, ACB_CTX_PREC(ctx));
+        if (!acb_is_finite(res))
+            return GR_UNABLE;
+    }
+    else
+    {
+        arb_floor(acb_realref(res), acb_realref(x), ACB_CTX_PREC(ctx));
+        arb_zero(acb_imagref(res));
+    }
+
     return GR_SUCCESS;
 }
 
 int
 _gr_acb_ceil(acb_t res, const acb_t x, const gr_ctx_t ctx)
 {
-    arb_ceil(acb_realref(res), acb_realref(x), ACB_CTX_PREC(ctx));
-    arb_zero(acb_imagref(res));
+    if (ACB_CTX_FLAGS(ctx) & ACB_CTX_ANALYTIC)
+    {
+        acb_real_ceil(res, x, 1, ACB_CTX_PREC(ctx));
+        if (!acb_is_finite(res))
+            return GR_UNABLE;
+    }
+    else
+    {
+        arb_ceil(acb_realref(res), acb_realref(x), ACB_CTX_PREC(ctx));
+        arb_zero(acb_imagref(res));
+    }
+
     return GR_SUCCESS;
 }
 
@@ -910,6 +971,152 @@ _gr_acb_arg(acb_t res, const acb_t x, const gr_ctx_t ctx)
     return GR_SUCCESS;
 }
 
+int
+_gr_acb_cmp(int * res, const acb_t x, const acb_t y, const gr_ctx_t ctx)
+{
+    if (arb_is_zero(acb_imagref(x)) && arb_is_zero(acb_imagref(y)) &&
+        ((arb_is_exact(acb_realref(x)) && arb_is_exact(acb_realref(y))) || !arb_overlaps(acb_realref(x), acb_realref(y))))
+    {
+        *res = arf_cmp(arb_midref(acb_realref(x)), arb_midref(acb_realref(y)));
+        return GR_SUCCESS;
+    }
+    else
+    {
+        *res = 0;
+        return GR_UNABLE;
+    }
+}
+
+int
+_gr_arb_cmpabs(int * res, const arb_t x, const arb_t y, const gr_ctx_t ctx);
+
+int
+_gr_acb_cmpabs(int * res, const acb_t x, const acb_t y, const gr_ctx_t ctx)
+{
+    if (arb_is_zero(acb_imagref(x)) && arb_is_zero(acb_imagref(y)))
+    {
+        arb_srcptr a = acb_realref(x);
+        arb_srcptr c = acb_realref(y);
+
+        /* OK; ignores the context object */
+        return _gr_arb_cmpabs(res, a, c, ctx);
+    }
+    else
+    {
+        slong prec = ACB_CTX_PREC(ctx);
+        int status = GR_SUCCESS;
+
+        arb_srcptr a = acb_realref(x);
+        arb_srcptr b = acb_imagref(x);
+        arb_srcptr c = acb_realref(y);
+        arb_srcptr d = acb_imagref(y);
+
+        mag_t xlo, xhi, ylo, yhi, t;
+
+        mag_init(xlo);
+        mag_init(xhi);
+        mag_init(ylo);
+        mag_init(yhi);
+        mag_init(t);
+
+        arb_get_mag_lower(xlo, a);
+        arb_get_mag_lower(t, b);
+        mag_mul_lower(xlo, xlo, xlo);
+        mag_mul_lower(t, t, t);
+        mag_add_lower(xlo, xlo, t);
+
+        arb_get_mag_lower(ylo, c);
+        arb_get_mag_lower(t, d);
+        mag_mul_lower(ylo, ylo, ylo);
+        mag_mul_lower(t, t, t);
+        mag_add_lower(ylo, ylo, t);
+
+        arb_get_mag(xhi, a);
+        arb_get_mag(t, b);
+        mag_mul(xhi, xhi, xhi);
+        mag_mul(t, t, t);
+        mag_add(xhi, xhi, t);
+
+        arb_get_mag(yhi, c);
+        arb_get_mag(t, d);
+        mag_mul(yhi, yhi, yhi);
+        mag_mul(t, t, t);
+        mag_add(yhi, yhi, t);
+
+        if (mag_cmp(xhi, ylo) < 0)
+        {
+            *res = -1;
+            status = GR_SUCCESS;
+        }
+        else if (mag_cmp(xlo, yhi) > 0)
+        {
+            *res = 1;
+            status = GR_SUCCESS;
+        }
+        else
+        {
+            arb_t t, u;
+
+            arb_init(t);
+            arb_init(u);
+
+            arb_mul(t, a, a, prec);
+            arb_addmul(t, b, b, prec);
+            arb_mul(u, c, c, prec);
+            arb_addmul(u, d, d, prec);
+
+            if ((arb_is_exact(t) && arb_is_exact(u)) || !arb_overlaps(t, u))
+            {
+                *res = arf_cmp(arb_midref(t), arb_midref(u));
+                status = GR_SUCCESS;
+            }
+            else
+            {
+                /* todo: worth it to do an exact computation? */
+                *res = 0;
+                status = GR_UNABLE;
+            }
+
+            arb_clear(t);
+            arb_clear(u);
+        }
+
+        mag_clear(xlo);
+        mag_clear(xhi);
+        mag_clear(ylo);
+        mag_clear(yhi);
+        mag_clear(t);
+
+        return status;
+    }
+}
+
+int
+_gr_acb_min(acb_t res, const acb_t x, const acb_t y, const gr_ctx_t ctx)
+{
+    if (arb_is_zero(acb_imagref(x)) && arb_is_zero(acb_imagref(y)))
+    {
+        arb_min(acb_realref(res), acb_realref(x), acb_realref(y), ACB_CTX_PREC(ctx));
+        arb_zero(acb_imagref(res));
+        return GR_SUCCESS;
+    }
+    else
+        return GR_UNABLE;
+}
+
+int
+_gr_acb_max(acb_t res, const acb_t x, const acb_t y, const gr_ctx_t ctx)
+{
+    if (arb_is_zero(acb_imagref(x)) && arb_is_zero(acb_imagref(y)))
+    {
+        arb_max(acb_realref(res), acb_realref(x), acb_realref(y), ACB_CTX_PREC(ctx));
+        arb_zero(acb_imagref(res));
+        return GR_SUCCESS;
+    }
+    else
+        return GR_UNABLE;
+}
+
 int
 _gr_acb_pi(acb_t res, const gr_ctx_t ctx)
 {
@@ -931,7 +1138,15 @@ _gr_acb_log(acb_t res, const acb_t x, const gr_ctx_t ctx)
         return GR_UNABLE;
     }
 
-    acb_log(res, x, ACB_CTX_PREC(ctx));
+    if (ACB_CTX_FLAGS(ctx) & ACB_CTX_ANALYTIC)
+    {
+        acb_log_analytic(res, x, 1, ACB_CTX_PREC(ctx));
+        if (!acb_is_finite(res))
+            return GR_UNABLE;
+    }
+    else
+        acb_log(res, x, ACB_CTX_PREC(ctx));
+
     return GR_SUCCESS;
 }
 
@@ -1228,13 +1443,7 @@ _gr_acb_gamma_fmpq(acb_t res, const fmpq_t x, const gr_ctx_t ctx)
     }
 }
 
-
-int
-_gr_acb_rgamma(acb_t res, const acb_t x, const gr_ctx_t ctx)
-{
-    acb_rgamma(res, x, ACB_CTX_PREC(ctx));
-    return GR_SUCCESS;
-}
+DEF_FUNC(rgamma)
 
 int
 _gr_acb_lgamma(acb_t res, const acb_t x, const gr_ctx_t ctx)
@@ -1540,12 +1749,7 @@ int _gr_acb_stieltjes(acb_t res, const fmpz_t n, const acb_t a, const gr_ctx_t c
     return acb_is_finite(res) ? GR_SUCCESS : GR_UNABLE;
 }
 
-int
-_gr_acb_dirichlet_eta(acb_t res, const acb_t x, const gr_ctx_t ctx)
-{
-    acb_dirichlet_eta(res, x, ACB_CTX_PREC(ctx));
-    return GR_SUCCESS;
-}
+DEF_FUNC(dirichlet_eta)
 
 /* todo
 int
@@ -2116,6 +2320,10 @@ gr_method_tab_input _acb_methods_input[] =
     {GR_METHOD_SGN,             (gr_funcptr) _gr_acb_sgn},
     {GR_METHOD_CSGN,            (gr_funcptr) _gr_acb_csgn},
     {GR_METHOD_ARG,             (gr_funcptr) _gr_acb_arg},
+    {GR_METHOD_CMP,             (gr_funcptr) _gr_acb_cmp},
+    {GR_METHOD_CMPABS,          (gr_funcptr) _gr_acb_cmpabs},
+    {GR_METHOD_MIN,             (gr_funcptr) _gr_acb_min},
+    {GR_METHOD_MAX,             (gr_funcptr) _gr_acb_max},
     {GR_METHOD_PI,              (gr_funcptr) _gr_acb_pi},
     {GR_METHOD_EXP,             (gr_funcptr) _gr_acb_exp},
     {GR_METHOD_EXPM1,           (gr_funcptr) _gr_acb_expm1},
@@ -2284,6 +2492,7 @@ gr_ctx_init_complex_acb(gr_ctx_t ctx, slong prec)
 
     ACB_CTX_PREC(ctx) = FLINT_MAX(2, FLINT_MIN(prec, WORD_MAX / 8));
     ACB_CTX_PREC(ctx) = prec;
+    ACB_CTX_FLAGS(ctx) = 0;
 
     ctx->methods = _acb_methods;
 
